#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import typing as t


def update_if_not_none(to_update: t.Dict[str, t.Any], new_values: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """
    Insert all values from ``new_values`` into ``to_update``, overwriting the values
    for any keys already present, unless the value is ``None``, in which case do nothing.

    Args:
        to_update (:obj:`dict`): dict object to be updated with additional values
        new_values (:obj:`dict`): dict object to be updated with additional values

    Returns:
        :obj:`dict`: to_update object updated with all values in new_values object which were not None
    """
    for k, v in new_values.items():
        if v is not None:
            to_update[k] = v
    return to_update



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

__all__ = ["User"]

import dataclasses
import datetime
import typing as t

from terra import exceptions
from terra.models.base_model import TerraDataModel

if t.TYPE_CHECKING:
    from terra import base_client
    from terra.api import api_responses


@dataclasses.dataclass
class User(TerraDataModel):
    def __init__(
        self,
        client: t.Optional[base_client.Terra] = None,
        user_id: t.Optional[str] = None,
        reference_id: t.Optional[str] = None,
        provider: t.Optional[str] = None,
        last_webhook_update: t.Optional[str] = None,
        scopes: t.Optional[str] = None,
    ) -> None:
        self.user_id = user_id
        self.reference_id = reference_id
        self.provider = provider
        self.last_webhook_update = last_webhook_update
        self.scopes = scopes
        self._client = client
        self._resource = None

    def _has_client(self) -> bool:
        return self._client is not None

    def fill_in_user_info(self) -> None:
        """
        Internal method used to retrieve data for User

        Args:
            dtype (:obj:`str`): datatype to be fetched


        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        if self._client:
            user_info = self._client.get_user_info(self)
            if not user_info.json:
                raise exceptions.NoUserInfoException

            self.provider = user_info.json["user"]["provider"]
            self.last_webhook_update = user_info.json["user"]["last_webhook_update"]
            self.scopes = user_info.json["user"]["scopes"]

    def _check_client(self) -> None:
        """
        Internal method used to check if user is connected to a client

        Returns:
            None
        """
        if self._client is None:
            raise exceptions.NoClientAvailable

    def get_activity(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves workouts/activity data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:

            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """

        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="activity",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_body(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves body metrics data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """

        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="body",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_nutrition(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves nutrition data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="nutrition",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_daily(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves daily summary data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="daily",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_sleep(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves sleep data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:

            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="sleep",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_athlete(
        self,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves profile info/athlete data for a given User object. By default, data will be asynchronously sent to
        registered webhook URL.

        Args:
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body
        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(dtype="athlete", user=self, to_webhook=to_webhook)

    def get_menstruation(
        self,
        start_date: datetime.datetime,
        end_date: t.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves daily summary data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed response object if no error has occured

        """
        if self._client is None:
            raise exceptions.NoClientAvailable

        return self._client._get_arbitrary_data(
            dtype="menstruation",
            user=self,
            start_date=int(start_date.timestamp()),
            end_date=int(end_date.timestamp()) if end_date is not None else None,
            to_webhook=to_webhook,
        )



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2 import samples as samples_
from terra.models.v2.activity import ActiveDurationsData
from terra.models.v2.activity import CaloriesData
from terra.models.v2.activity import DeviceData
from terra.models.v2.activity import ElevationSummary
from terra.models.v2.activity import HeartRateData
from terra.models.v2.activity import METData
from terra.models.v2.activity import OxygenData
from terra.models.v2.activity import SwimmingSummary

__all__ = ["Daily"]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    upload_type: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class DistanceDataDetailed(base_model.TerraDataModel):
    step_samples: typing.List[samples_.StepSample] = dataclasses.field(default_factory=list)
    distance_samples: typing.List[samples_.DistanceSample] = dataclasses.field(default_factory=list)
    elevation_samples: typing.List[samples_.ElevationSample] = dataclasses.field(default_factory=list)
    floors_climbed_samples: typing.List[samples_.FloorsClimbedSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class StrainData(base_model.TerraDataModel):
    strain_level: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class DistanceData(base_model.TerraDataModel):
    distance_meters: typing.Optional[float] = dataclasses.field(default=None)
    steps: typing.Optional[int] = dataclasses.field(default=None)
    floors_climbed: typing.Optional[int] = dataclasses.field(default=None)
    swimming: SwimmingSummary = dataclasses.field(default_factory=SwimmingSummary)
    elevation: ElevationSummary = dataclasses.field(default_factory=ElevationSummary)
    detailed: DistanceDataDetailed = dataclasses.field(default_factory=DistanceDataDetailed)


@dataclasses.dataclass
class TagEntry(base_model.TerraDataModel):
    tag_name: typing.Optional[str] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    notes: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class TagData(base_model.TerraDataModel):
    tags: typing.List[TagEntry] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class StressData(base_model.TerraDataModel):
    stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    rest_stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    activity_stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    low_stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    medium_stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    high_stress_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    avg_stress_level: typing.Optional[float] = dataclasses.field(default=None)
    max_stress_level: typing.Optional[float] = dataclasses.field(default=None)
    samples: typing.List[samples_.StressSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Scores(base_model.TerraDataModel):
    recovery: typing.Optional[float] = dataclasses.field(default=None)
    activity: typing.Optional[float] = dataclasses.field(default=None)
    sleep: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class Daily(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    tag_data: TagData = dataclasses.field(default_factory=TagData)
    active_durations_data: ActiveDurationsData = dataclasses.field(default_factory=ActiveDurationsData)
    distance_data: DistanceData = dataclasses.field(default_factory=DistanceData)
    heart_rate_data: HeartRateData = dataclasses.field(default_factory=HeartRateData)
    calories_data: CaloriesData = dataclasses.field(default_factory=CaloriesData)
    MET_data: METData = dataclasses.field(default_factory=METData)
    stress_data: StressData = dataclasses.field(default_factory=StressData)
    oxygen_data: OxygenData = dataclasses.field(default_factory=OxygenData)
    strain_data: StrainData = dataclasses.field(default_factory=StrainData)
    device_data: DeviceData = dataclasses.field(default_factory=DeviceData)
    scores: Scores = dataclasses.field(default_factory=Scores)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

__all__ = ["TerraDataModel"]

import dataclasses
import enum
import pydoc
import typing

PRIMITIVES = (str, int, bool, float, type(None), dict)
datamodelT = typing.TypeVar("datamodelT", bound="TerraDataModel")


class TerraDataModel:
    """
    Base class for all data models that terra returns.
    """

    def _get_attrs(self) -> typing.Iterable[str]:
        return filter(
            lambda a: not a.startswith("_"),
            set(dir(self)).difference(set(dir(TerraDataModel))),
        )

    def keys(self) -> typing.Generator[str, None, None]:
        yield from self._get_attrs()

    def values(self) -> typing.Generator[typing.Any, None, None]:
        attrs = self._get_attrs()
        yield from (getattr(self, attr) for attr in attrs)

    def items(self) -> typing.Generator[typing.Tuple[str, typing.Any], None, None]:
        attrs = self._get_attrs()
        for attr in attrs:
            yield attr, getattr(self, attr)

    def to_dict(self) -> typing.Dict[str, typing.Any]:
        """
        Get the dictionary (json) representation of the data model.

        This method inspects the attributes of the instance that it is being called on
        to determine how to build the correct payload from the data stored.

        Returns:
            :obj:`dict`: Dictionary representation of the data model.
        """
        attrs = self._get_attrs()

        output: typing.Dict[typing.Any, typing.Any] = {}
        for attr in attrs:
            attr_val = getattr(self, attr)
            if isinstance(attr_val, enum.IntEnum):
                output[attr] = int(attr_val)
            elif type(attr_val) in PRIMITIVES:
                output[attr] = attr_val
            elif isinstance(attr_val, list):
                if (attr_val and type(attr_val[0]) in PRIMITIVES) or not attr_val:
                    output[attr] = attr_val
                else:
                    output[attr] = [item.to_dict() for item in attr_val]
            else:
                output[attr] = attr_val.to_dict()
        return output

    def filter_data(self: TerraDataModel, term: str) -> typing.Generator[datamodelT, None, None]:
        """
        Returns a generator of all the data models that match the filter

        Args:
            term:obj:`str`: the word to filter with


        Returns:
            :obj:`typing.Generator[datamodelT]`
        """

        fields_dict = {field.name: field.type for field in dataclasses.fields(self)}
        # print(fields_dict)

        for field_name, field_type in fields_dict.items():
            try:
                if isinstance(getattr(self, field_name, None), TerraDataModel):
                    for sub_term in field_name.lower().split("_"):
                        if sub_term.lower() == term.lower():
                            yield typing.cast(datamodelT, getattr(self, field_name, None))
                    # print(getattr(self, field_name , None))
                    typing.cast(datamodelT, getattr(self, field_name, None)).filter_data(term)

            except Exception as e:
                try:
                    for sub_term in field_name.lower().split("_"):
                        if sub_term.lower() == term.lower():
                            yield typing.cast(datamodelT, getattr(self, field_name, None))

                    for inner_item in typing.cast(typing.List[TerraDataModel], getattr(self, field_name, None)):
                        # print(inner_item)
                        inner_item.filter_data(term)

                except Exception as e:
                    # print(traceback.format_exc())
                    pass

    @classmethod
    def from_dict(
        cls: typing.Type[datamodelT], model_dict: typing.Dict[str, typing.Any], safe: bool = True
    ) -> datamodelT:
        """
        Return the Class data model representation of the dictionary (json).

        This method inspects the attributes of the class that it is being called on
        to determine how to build the correct payload from the data stored.

        Args:
            model_dict:obj:`dict`:
            safe:obj:`bool`:

        Returns:
            :obj:`terrpython.models.base_model.TerraDataModel`
        """
        # TODO - I don't like this function at all. It can definitely be more elegant
        data_model = cls()
        for k, v in model_dict.items():
            if (
                (inner_item := getattr(data_model, k, *(("NOT_FOUND",) if safe else ()))) in [None, []]
                or isinstance(inner_item, TerraDataModel)
                or isinstance(v, list)
            ):
                if isinstance(inner_item, TerraDataModel):
                    v = inner_item.from_dict(v)

                # if it's a list
                if isinstance(v, list):
                    if v != []:
                        # getting all the field types of the current class
                        fields_dict = {field.name: field.type for field in dataclasses.fields(cls())}

                        # getting the current field type as a string and removing the 't.optional'
                        current_field_type = str(fields_dict[k]).split("[")[1].split("]")[0]
                        current_field_type2 = str(fields_dict[k]).split("[")[1].split("]")[0]

                        # adding terra before the models name
                        if current_field_type.split(".")[0] == "models":
                            current_field_type = "terra." + current_field_type

                        if current_field_type2.split(".")[0] == "models":
                            current_field_type2 = "terra-client." + current_field_type2

                        # check if the elements of the list are Terra Data Models
                        if current_field_type.split(".")[0] == "terra":
                            result = []

                            # for each json object inside the list
                            for inner_dict in v:
                                # an instance of a data model of the type of items inside the list
                                inner_data_model = typing.cast(
                                    typing.Type[TerraDataModel], pydoc.locate(current_field_type)
                                )()

                                # fill up the model
                                inner_data_model = inner_data_model.from_dict(inner_dict)

                                # append the model to the result list
                                result.append(inner_data_model)

                            v = result

                setattr(data_model, k, v)

        return data_model

    def populate_from_dict(
        self: datamodelT, model_dict: typing.Dict[str, typing.Any], safe: bool = False
    ) -> datamodelT:
        """
        Populates missing data fields in the class given a dictionary (json).

        This method inspects the attributes of the instance that it is being called on
        to determine how to build the correct payload from the data stored.

        Args:
            model_dict:obj:`dict`:
            safe:obj:`bool`:

        Returns:
            :obj:`terrpython.models.base_model.TerraDataModel`
        """
        for k, v in model_dict.items():
            if getattr(self, k, *(("NOT_FOUND",) if safe else ())) is None:
                setattr(self, k, v)

        return self



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model

__all__ = ["ActivitySample"]


@dataclasses.dataclass
class ActivitySample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    active_duration_seconds: typing.Optional[int] = dataclasses.field(default=None)
    altitude_in_meters: typing.Optional[float] = dataclasses.field(default=None)
    calories: typing.Optional[float] = dataclasses.field(default=None)
    distance_in_km: typing.Optional[float] = dataclasses.field(default=None)
    heartrate_bpm: typing.Optional[float] = dataclasses.field(default=None)
    coordinates_lat_lng: typing.List[float] = dataclasses.field(default_factory=list)
    rep_count: typing.Optional[int] = dataclasses.field(default=None)
    resting_duration_seconds: typing.Optional[int] = dataclasses.field(default=None)
    speed_in_meters_per_seconds: typing.Optional[float] = dataclasses.field(default=None)
    spo2_percentage: typing.Optional[float] = dataclasses.field(default=None)
    status: typing.Optional[str] = dataclasses.field(default=None)
    swimming_laps: typing.Optional[float] = dataclasses.field(default=None)
    swimming_strokes: typing.Optional[int] = dataclasses.field(default=None)
    vo2_volume_ml_per_min_per_kg: typing.Optional[float] = dataclasses.field(default=None)
    vo2_max_volume_ml_per_min_per_kg: typing.Optional[float] = dataclasses.field(default=None)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

__all__ = ["Terra"]

import datetime
import hashlib
import hmac
import json
import typing

import requests

from terra import constants
from terra import utils
from terra.api import api_responses
from terra.models import user as user_

if typing.TYPE_CHECKING:
    import flask


class Terra:
    """
    constructor of the Terra class

    Args:
        api_key (:obj:`str`) : Your API Key
        dev_id (:obj:`str`) : Your dev ID
        secret (:obj:`str`) : Your terra secret (for web hooks)

    """

    def __init__(self, api_key: str, dev_id: str, secret: str) -> None:
        self.api_key = api_key
        self.dev_id = dev_id
        self.secret = secret

    @property
    def _auth_headers(self) -> typing.Dict[str, str]:
        """
        Internal method used to fill in authentication headers for all requests to the API

        Returns:
            :obj:`dict`: Dictionary of required auth headers

        """
        return {"x-api-key": self.api_key, "dev-id": self.dev_id}

    def from_user_id(self, user_id: str) -> user_.User:
        """
        Creates a User instance out of a UUID corresponding to a registered User on the API

        Args:
            user_id (:obj:`str`): UUID corresponding to a user currently authenticated on the API

        Returns:
            :obj:`User`: Created User instance

        """

        user = user_.User(user_id=user_id, client=self)
        user.fill_in_user_info()
        return user

    def _get_arbitrary_data(self, user: user_.User, dtype: str, **kwargs: typing.Any) -> api_responses.TerraApiResponse:
        """
        Internal method used to retrieve data for a given User

        Args:
            user (:obj:`models.user.User`):
            dtype (:obj:`str`): datatype to be fetched
            **kwargs: optional additional parameters for the request

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        params = {"user_id": user.user_id}
        params = utils.update_if_not_none(params, kwargs)

        data_resp = requests.get(
            f"{constants.BASE_URL}/{dtype}",
            params=params,
            headers=self._auth_headers,
        )

        return api_responses.TerraApiResponse(data_resp, user=user, dtype=dtype)

    def get_activity_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves workouts/activity data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will
                default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        return user.get_activity(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_body_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves body metrics data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will
                default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        return user.get_body(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_daily_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves daily summary data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will
                default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        return user.get_daily(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_sleep_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves sleep data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will
                default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        return user.get_sleep(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_athlete_for_user(
        self,
        user: user_.User,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves profile info/athlete data for a given User object. By default, data will be asynchronously sent to
        registered webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """
        return self._get_arbitrary_data(user, "athlete", to_webhook=to_webhook)

    def get_menstruation_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves daily summary data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set,
                will default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """

        return user.get_menstruation(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def get_nutrition_for_user(
        self,
        user: user_.User,
        start_date: datetime.datetime,
        end_date: typing.Optional[datetime.datetime] = None,
        to_webhook: bool = True,
    ) -> api_responses.TerraApiResponse:
        """
        Retrieves daily summary data for a given User object. By default, data will be asynchronously sent to registered
        webhook URL.

        Args:
            user (:obj:`models.user.User`): User for whom to fetch data
            start_date (:obj:`datetime.datetime`): Datetime object for which to fetch data
            end_date:obj (:`datetime.datetime`): Optional end_date for which to fetch data - if not set, will
                default to start_date + 24h according to current API specifications
            to_webhook (:obj:`bool`): Whether to send data to registered webhook URL or return as a response body

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing DataReturned parsed
                response object if no error has occured

        """

        return user.get_nutrition(
            start_date=start_date,
            end_date=end_date if end_date is not None else None,
            to_webhook=to_webhook,
        )

    def generate_widget_session(
        self,
        providers: typing.List[str],
        auth_success_redirect_url: typing.Optional[str] = None,
        auth_failure_redirect_url: typing.Optional[str] = None,
        language: typing.Optional[str] = None,
        reference_id: typing.Optional[str] = None,
        **kwargs: typing.Any,
    ) -> api_responses.TerraApiResponse:
        """
        Generates a widget session used to allow an end user to authenticate through the API. Users should be
        redirected to the given URL in order to complete authentication

        Args:
            providers (List[:obj:`str`]): Providers to display on widget wearable selection screen, by leaving it empty it will use all default providers
            auth_success_redirect_url (Optional[:obj:`str`]): URL to redirect to upon successful authentication
            auth_failure_redirect_url (Optional[:obj:`str`]): URL to redirect to upon unsuccessful authentication
            language (Optional[:obj:`str`]): Language to display widget in
            reference_id (Optional[:obj:`str`]): ID of a user in your app, which will be returned at the end of a successful auth
            **kwargs: Optional additional arguments to be passed in to the body of the request

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing WidgetSession parsed response object if no error has occured
        """
        maybe_body_payload = {
            "providers": ",".join(providers) if providers else None,
            "auth_success_redirect_url": auth_success_redirect_url,
            "auth_failure_redirect_url": auth_failure_redirect_url,
            "language": language,
            "reference_id": reference_id,
        }
        body_payload = utils.update_if_not_none({}, maybe_body_payload)
        body_payload.update(kwargs)

        widget_resp = requests.post(
            f"{constants.BASE_URL}/auth/generateWidgetSession",
            headers=self._auth_headers,
            json=body_payload,
        )
        return api_responses.TerraApiResponse(widget_resp, dtype="widget_session")

    def generate_authentication_url(
        self,
        resource: str,
        auth_success_redirect_url: typing.Optional[str] = None,
        auth_failure_redirect_url: typing.Optional[str] = None,
        reference_id: typing.Optional[str] = None,
        **kwargs: typing.Any,
    ) -> api_responses.TerraApiResponse:
        """
        Generates an authentication URL to allow an end user to authenticate through the API. Users should be
        redirected to the given URL in order to complete authentication. User ID will be provided in the response
        for convenience (note that at this stage, said user will have yet to complete the auth flow)

        Args:
            resource (:obj:`str`): Provider to authenticate user with
            auth_success_redirect_url (Optional[:obj:`str`]): URL to redirect to upon successful authentication
            auth_failure_redirect_url (Optional[:obj:`str`]): URL to redirect to upon unsuccessful authentication
            reference_id (Optional[:obj:`str`]): ID of a user in your app, which will be returned at the
                end of a successful auth
            **kwargs: Optional additional arguments to be passed in to the body of the request

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing UserAuthUrl parsed
                response object if no error has occured

        """

        body_payload = {
            "resource": resource,
            "auth_success_redirect_url": auth_success_redirect_url,
            "auth_failure_redirect_url": auth_failure_redirect_url,
            "reference_id": reference_id,
        }
        body_payload.update(kwargs)

        auth_resp = requests.post(
            f"{constants.BASE_URL}/auth/authenticateUser",
            headers=self._auth_headers,
            json=body_payload,
        )

        return api_responses.TerraApiResponse(auth_resp, dtype="auth_url")

    def get_user_info(self, user: user_.User) -> api_responses.TerraApiResponse:
        """
        Retrieve information on a given User, including is_authenticated status, indicating if the user has
        successfully completed auth flow, or has yet to do so
        Note: Also updates information on user object passed as an argument

        Args:
            user (:obj:`models.user.User`): User to retrieve information for

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing UserInfo parsed
                response object if no error has occured
        """
        user_resp = requests.get(
            f"{constants.BASE_URL}/userInfo",
            params={"user_id": user.user_id},
            headers=self._auth_headers,
        )
        return api_responses.TerraApiResponse(user_resp, dtype="user_info")

    def deauthenticate_user(self, user: user_.User) -> api_responses.TerraApiResponse:
        """
        Deauthenticates the given User from the Api. If successful, this will trigger a `deauth`
        webhook event.

        Args:
            user (:obj:`models.user.User`): User to Deauthenticate from the API

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing UserDeauthResp parsed response object if no error has occured
        """
        deauth_resp = requests.delete(
            f"{constants.BASE_URL}/auth/deauthenticateUser",
            params={"user_id": user.user_id},
            headers=self._auth_headers,
        )
        deauth_resp.raise_for_status()
        return api_responses.TerraApiResponse(deauth_resp, dtype="user_deauth")

    def list_users(self) -> api_responses.TerraApiResponse:
        """
        Lists all users registered under Client's credentials on the API

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing SubscribedUsers parsed response object if no error has occured
        """
        users_resp = requests.get(f"{constants.BASE_URL}/subscriptions", headers=self._auth_headers)
        return api_responses.TerraApiResponse(users_resp, dtype="subscriptions", client=self)

    def list_providers(self) -> api_responses.TerraApiResponse:
        """
        Lists all providers on the API

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing ProvidersResponse parsed response object if no error has occured
        """
        providers_resp = requests.get(f"{constants.BASE_URL}/integrations", headers=self._auth_headers)
        return api_responses.TerraApiResponse(providers_resp, dtype="providers")

    def check_terra_signature(self, body: str, header: str) -> bool:
        """
        Function to test if the body of an API response comes from terra using SHA256

        Args:
            body (:obj:`str`): The body from API response as a string
            header (:obj:`str`): The header from API response as a string

        Returns:
            :obj:`bool`: True if the API response comes from Terra
        """

        t, sig = (pair.split("=")[-1] for pair in header.split(","))

        computed_signature = hmac.new(
            bytes(self.secret, "utf-8"),
            msg=bytes(f"{t}.{body}", "utf-8"),
            digestmod=hashlib.sha256,
        ).hexdigest()

        if computed_signature != sig:
            return False
        # Signature was validated
        return True

    def handle_flask_webhook(self, request: flask.Request) -> typing.Optional[api_responses.TerraWebhookResponse]:
        """
        Parses Terra webhooks from a flask request

        Args:
            request (:obj:`flask.request`): the flask request object

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing ProvidersResponse parsed
                response object if no error has occurred
        """

        if not self.check_terra_signature(request.get_data().decode("utf-8"), request.headers["terra-signature"]):
            return None
        ff = api_responses.TerraWebhookResponse(request.get_json(), dtype="hook")

        return ff

    def handle_webhook(
        self, payload: str, terra_signature_header: str
    ) -> typing.Optional[api_responses.TerraWebhookResponse]:
        """
        Function to Parse web hooks from Terra

        Args:
            payload (:obj:`str`): The body from API response as a string
            terra_signature_header (:obj:`str`): The terra_signature header from API response as a string

        Returns:
            :obj:`models.api_responses.TerraApiResponse`: API response object containing ProvidersResponse parsed
                response object if no error has occurred
        """

        if not self.check_terra_signature(payload, terra_signature_header):
            return None
        return api_responses.TerraWebhookResponse(json.loads(payload), dtype="hook")



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

__all__ = [
    "NoDataReturned",
    "DataReturned",
    "AuthenticationFailed",
    "TerraParsedApiResponse",
    "TerraWebhookResponse",
    "TerraApiResponse",
    "GenericMessage",
    "GoogleNoDataSourceHookResponse",
    "WidgetSession",
    "SentToWebhook",
    "SubscribedUsers",
    "UserDeauthResp",
    "AuthHookResponse",
    "AccessRevokedHookResponse",
    "UserAuthUrl",
    "UserDeauthHookResponse",
    "UserReauthHookResponse",
    "ConnectionDegraded",
    "ProvidersResponse",
    "RequestProcessingHookResponse",
    "ConnectionErrorHookResponse",
    "HookResponse",
    "RequestCompletedHookResponse",
    "UserInfo",
]

import dataclasses
import json.decoder
import typing

import requests

from terra import exceptions
from terra import models
from terra.models import base_model
from terra.models import user as user_

if typing.TYPE_CHECKING:
    from terra.base_client import Terra


# TODO - should use a mixin/trait here instead of a redundant subclass probably
class TerraParsedApiResponse(base_model.TerraDataModel):
    pass


def _parse_api_body(
    dtype: typing.Optional[str],
    body: typing.Optional[typing.Dict[str, typing.Any]],
    user: typing.Optional[models.user.User],
) -> TerraParsedApiResponse:
    if not body:
        raise exceptions.NoBodyException

    Auser = user
    if "user" in body:
        Auser = models.user.User.from_dict(body["user"])

    if ("status" in body) and (body["status"] in STATUS.keys()):
        response = STATUS[body["status"]]().from_dict(body)

    elif dtype in USER_DATATYPES:
        if not dtype:
            raise exceptions.NoDtypeException

        return DataReturned(
            user=Auser,
            data=[MODEL_MAPPING[dtype]().from_dict(item) for item in body["data"]]
            if body.get("data") or body.get("data") == []
            else [],
            type=dtype,
        )
    elif dtype in DTYPE_TO_RESPONSE.keys():
        if not dtype:
            raise exceptions.NoDtypeException

        response = DTYPE_TO_RESPONSE[dtype]().from_dict(body)

    elif dtype in HOOK_RESPONSE.keys():
        if not dtype:
            raise exceptions.NoDtypeException
        response = HOOK_RESPONSE[dtype]().from_dict(body)

    else:
        response = GenericMessage().from_dict(body)

    try:
        setattr(response, "user", Auser)

    finally:
        try:
            if "old_user" in body:
                setattr(
                    response,
                    "old_user",
                    models.user.User.from_dict(body["old_user"]),
                )
            if "new_user" in body:
                setattr(
                    response,
                    "new_user",
                    models.user.User.from_dict(body["new_user"]),
                )
        finally:
            return typing.cast(TerraParsedApiResponse, response)


class TerraApiResponse(TerraParsedApiResponse):
    def __init__(
        self,
        resp: requests.Response,
        dtype: str,
        user: typing.Optional[user_.User] = None,
        client: typing.Optional[Terra] = None,
    ) -> None:
        self.response_code: int = resp.status_code
        self.raw_body: typing.Optional[str] = resp.text
        self.json: typing.Optional[typing.Dict[str, typing.Any]] = None
        self.dtype: typing.Optional[str] = dtype
        try:
            body: typing.Dict[str, typing.Any] = resp.json()
            self.json = body
            self.dtype = body.get("type", dtype)
            self.parsed_response: TerraParsedApiResponse = _parse_api_body(self.dtype, body, user)
            if client:
                for e in typing.cast(SubscribedUsers, self.parsed_response).users:
                    e._client = client
        except json.decoder.JSONDecodeError:
            resp.raise_for_status()

    def get_parsed_response(self) -> TerraParsedApiResponse:
        """
        Get the parsed representation of the api response.

        Returns:
            :obj:`TerraParsedApiResponse`: Parsed representation of the data model.
        """
        return self.parsed_response

    def get_json(self) -> typing.Optional[typing.Dict[str, typing.Any]]:
        """
        Get the dictionary (json) representation of the api response.

        Returns:
            :obj:`dict`: Dictionary representation of the api response.
        """
        return self.json


class TerraWebhookResponse(TerraParsedApiResponse):
    def __init__(
        self,
        resp: typing.Union[typing.Dict[str, typing.Any], requests.Response],
        user: typing.Optional[user_.User] = None,
        dtype: typing.Optional[str] = None,
    ) -> None:
        self.json: typing.Dict[str, typing.Any] = resp.json() if isinstance(resp, requests.Response) else resp
        self.dtype: typing.Optional[str] = self.json.get("type", dtype)
        self.parsed_response: TerraParsedApiResponse = _parse_api_body(self.dtype, self.json, user)

    def get_parsed_response(self) -> TerraParsedApiResponse:
        """
        Get the parsed representation of the api response.

        Returns:
            :obj:`TerraParsedApiResponse`: Parsed representation of the data model.
        """
        return self.parsed_response

    def get_json(self) -> typing.Dict[str, typing.Any]:
        """
        Get the dictionary (json) representation of the api response.

        Returns:
            :obj:`dict`: Dictionary representation of the api response.
        """
        return self.json


@dataclasses.dataclass
class GenericMessage(TerraParsedApiResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    status: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class WidgetSession(TerraParsedApiResponse):
    expires_in: int = dataclasses.field(default=900)
    status: typing.Optional[str] = dataclasses.field(default=None)
    session_id: typing.Optional[str] = dataclasses.field(default=None)
    url: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class UserInfo(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)
    is_authenticated: bool = dataclasses.field(default=True)


@dataclasses.dataclass
class UserDeauthResp(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default="success")


@dataclasses.dataclass
class HookResponse(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default="success")
    type: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class AuthHookResponse(HookResponse):
    reference_id: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)
    widget_session_id: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class UserReauthHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    old_user: typing.Optional[models.user.User] = dataclasses.field(default=None)
    new_user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class UserDeauthHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class AccessRevokedHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class GoogleNoDataSourceHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class ConnectionErrorHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class RequestProcessingHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class RequestCompletedHookResponse(HookResponse):
    message: typing.Optional[str] = dataclasses.field(default=None)
    reference: typing.Optional[str] = dataclasses.field(default=None)
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)


@dataclasses.dataclass
class SubscribedUsers(TerraParsedApiResponse):
    users: typing.List[models.user.User] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class UserAuthUrl(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default=None)
    expires_in: int = dataclasses.field(default=900)
    auth_url: typing.Optional[str] = dataclasses.field(default=None)
    user_id: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class NoDataReturned(TerraParsedApiResponse):
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)
    status: typing.Optional[str] = dataclasses.field(default="not_available")
    message: typing.Optional[str] = dataclasses.field(default="Data type requested not available from provider")


@dataclasses.dataclass
class DataReturned(TerraParsedApiResponse):
    user: typing.Optional[models.user.User] = dataclasses.field(default=None)
    type: typing.Optional[str] = dataclasses.field(default=None)
    data: typing.List[TerraParsedApiResponse] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class AuthenticationFailed(TerraParsedApiResponse):
    status: str = dataclasses.field(default="error")
    widget_session_id: typing.Optional[str] = dataclasses.field(default=None)
    reference_id: typing.Optional[str] = dataclasses.field(default=None)
    message: typing.Optional[str] = dataclasses.field(default="User failed to authenticate and has been deleted")
    type: str = dataclasses.field(default="auth")
    reason: str = dataclasses.field(default="auth_cancelled")


@dataclasses.dataclass
class ConnectionDegraded(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default="warning")
    message: typing.Optional[str] = dataclasses.field(default="User connection degraded")
    type: typing.Optional[str] = dataclasses.field(default="connection_error")


@dataclasses.dataclass
class ProvidersResponse(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default="warning")
    providers: typing.Optional[typing.List[str]] = dataclasses.field(default=None)
    sdk_providers: typing.Optional[typing.List[str]] = dataclasses.field(default=None)


@dataclasses.dataclass
class SentToWebhook(TerraParsedApiResponse):
    status: typing.Optional[str] = dataclasses.field(default=None)
    message: typing.Optional[str] = dataclasses.field(default=None)


USER_DATATYPES = [
    "activity",
    "athlete",
    "body",
    "daily",
    "menstruation",
    "sleep",
    "nutrition",
]
MODEL_MAPPING = {
    "activity": models.v2.activity.Activity,
    "body": models.v2.body.Body,
    "daily": models.v2.daily.Daily,
    "sleep": models.v2.sleep.Sleep,
    "menstruation": models.v2.menstruation.Menstruation,
    "athlete": models.v2.athlete.Athlete,
    "nutrition": models.v2.nutrition.Nutrition,
}


DTYPE_TO_RESPONSE = {
    "widget_session": WidgetSession,
    "auth_url": UserAuthUrl,
    "user_info": UserInfo,
    "subscriptions": SubscribedUsers,
    "providers": ProvidersResponse,
    "sent_to_webhook": SentToWebhook,
}

HOOK_TYPES = {
    "auth",
    "user_reauth",
    "access_revoked",
    "deauth",
    "google_no_datasource",
    "connexion_error",
    "request_processing",
    "request_completed",
}

HOOK_RESPONSE = {
    "auth": AuthHookResponse,
    "user_reauth": UserReauthHookResponse,
    "access_revoked": AccessRevokedHookResponse,
    "deauth": UserDeauthHookResponse,
    "google_no_datasource": GoogleNoDataSourceHookResponse,
    "connexion_error": ConnectionErrorHookResponse,
    "request_completed": RequestCompletedHookResponse,
    "request_processing": RequestProcessingHookResponse,
}

STATUS = {
    "not_available": NoDataReturned,
    "warning": ConnectionDegraded,
    "error": ConnectionDegraded,
}



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2 import samples as samples_
from terra.models.v2.activity import DeviceData

__all__ = [
    "Sleep",
    "Metadata",
    "ReadinessData",
    "TemperatureData",
    "HeartRateDataSummary",
    "HeartRateDataDetailed",
    "HeartRateData",
    "SleepDurationsAwakeData",
    "SleepDurationsAsleepData",
    "SleepDurationsOtherData",
    "SleepDurationsData",
    "OxygenSaturationData",
    "BreathsData",
    "SnoringData",
    "RespirationData",
]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    upload_type: typing.Optional[int] = dataclasses.field(default=None)
    is_nap: bool = dataclasses.field(default=False)


@dataclasses.dataclass
class ReadinessData(base_model.TerraDataModel):
    readiness: typing.Optional[float] = dataclasses.field(default=None)
    recovery_level: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class TemperatureData(base_model.TerraDataModel):
    delta: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartRateDataSummary(base_model.TerraDataModel):
    avg_hr_bpm: typing.Optional[int] = dataclasses.field(default=None)
    max_hr_bpm: typing.Optional[int] = dataclasses.field(default=None)
    min_hr_bpm: typing.Optional[int] = dataclasses.field(default=None)
    avg_hrv_rmssd: typing.Optional[int] = dataclasses.field(default=None)
    avg_hrv_sdnn: typing.Optional[int] = dataclasses.field(default=None)
    user_max_hr_bpm: typing.Optional[int] = dataclasses.field(default=None)
    resting_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartRateDataDetailed(base_model.TerraDataModel):
    hr_samples: typing.List[samples_.HeartRateDataSample] = dataclasses.field(default_factory=list)
    hrv_samples_rmssd: typing.List[samples_.HeartRateVariabilityDataSampleRMSSD] = dataclasses.field(
        default_factory=list
    )
    hrv_samples_sdnn: typing.List[samples_.HeartRateVariabilityDataSampleSDNN] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class HeartRateData(base_model.TerraDataModel):
    summary: HeartRateDataSummary = dataclasses.field(default_factory=HeartRateDataSummary)
    detailed: HeartRateDataDetailed = dataclasses.field(default_factory=HeartRateDataDetailed)


@dataclasses.dataclass
class SleepDurationsAwakeData(base_model.TerraDataModel):
    sleep_latency_seconds: typing.Optional[float] = dataclasses.field(default=None)
    wake_up_latency_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_awake_state_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_short_interruption_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_long_interruption_seconds: typing.Optional[float] = dataclasses.field(default=None)
    num_out_of_bed_events: typing.Optional[int] = dataclasses.field(default=None)
    num_wakeup_events: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class SleepDurationsAsleepData(base_model.TerraDataModel):
    duration_asleep_state_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_deep_sleep_state_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_light_sleep_state_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_REM_sleep_state_seconds: typing.Optional[float] = dataclasses.field(default=None)
    num_REM_events: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class SleepDurationsOtherData(base_model.TerraDataModel):
    duration_unmeasurable_sleep_seconds: typing.Optional[float] = dataclasses.field(default=None)
    duration_in_bed_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SleepDurationsData(base_model.TerraDataModel):
    sleep_efficiency: typing.Optional[float] = dataclasses.field(default=None)
    awake: SleepDurationsAwakeData = dataclasses.field(default_factory=SleepDurationsAwakeData)
    asleep: SleepDurationsAsleepData = dataclasses.field(default_factory=SleepDurationsAsleepData)
    other: SleepDurationsOtherData = dataclasses.field(default_factory=SleepDurationsOtherData)
    hypnogram_samples: typing.List[samples_.SleepHypnogramSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class OxygenSaturationData(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    avg_saturation_percentage: typing.Optional[float] = dataclasses.field(default=None)
    samples: typing.List[samples_.OxygenSaturationSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class BreathsData(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    on_demand_reading: typing.Optional[bool] = dataclasses.field(default=None)
    avg_breaths_per_min: typing.Optional[float] = dataclasses.field(default=None)
    max_breaths_per_min: typing.Optional[float] = dataclasses.field(default=None)
    min_breaths_per_min: typing.Optional[float] = dataclasses.field(default=None)
    samples: typing.List[samples_.BreathSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class SnoringData(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    total_snoring_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)
    num_snoring_events: typing.Optional[int] = dataclasses.field(default=None)
    samples: typing.List[samples_.SnoringSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class RespirationData(base_model.TerraDataModel):
    oxygen_saturation_data: OxygenSaturationData = dataclasses.field(default_factory=OxygenSaturationData)
    breaths_data: BreathsData = dataclasses.field(default_factory=BreathsData)
    snoring_data: SnoringData = dataclasses.field(default_factory=SnoringData)


@dataclasses.dataclass
class Sleep(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    temperature_data: TemperatureData = dataclasses.field(default_factory=TemperatureData)
    readiness_data: ReadinessData = dataclasses.field(default_factory=ReadinessData)
    heart_rate_data: HeartRateData = dataclasses.field(default_factory=HeartRateData)
    sleep_durations_data: SleepDurationsData = dataclasses.field(default_factory=SleepDurationsData)
    respiration_data: RespirationData = dataclasses.field(default_factory=RespirationData)
    device_data: DeviceData = dataclasses.field(default_factory=DeviceData)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2 import samples as samples_

__all__ = ["Activity"]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    name: typing.Optional[str] = dataclasses.field(default=None)
    type: typing.Optional[int] = dataclasses.field(default=None)
    summary_id: typing.Optional[str] = dataclasses.field(default=None)
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    city: typing.Optional[str] = dataclasses.field(default=None)
    state: typing.Optional[str] = dataclasses.field(default=None)
    country: typing.Optional[str] = dataclasses.field(default=None)
    upload_type: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class LapData(base_model.TerraDataModel):
    laps: typing.List[samples_.LapSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class StrainData(base_model.TerraDataModel):
    strain_level: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SwimmingSummary(base_model.TerraDataModel):
    num_laps: typing.Optional[int] = dataclasses.field(default=None)
    num_strokes: typing.Optional[int] = dataclasses.field(default=None)
    pool_length_meters: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class ElevationSummary(base_model.TerraDataModel):
    gain_planned_meters: typing.Optional[float] = dataclasses.field(default=None)
    gain_actual_meters: typing.Optional[float] = dataclasses.field(default=None)
    loss_actual_meters: typing.Optional[float] = dataclasses.field(default=None)
    min_meters: typing.Optional[float] = dataclasses.field(default=None)
    avg_meters: typing.Optional[float] = dataclasses.field(default=None)
    max_meters: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class DistanceDataSummary(base_model.TerraDataModel):
    distance_meters: typing.Optional[float] = dataclasses.field(default=None)
    steps: typing.Optional[int] = dataclasses.field(default=None)
    floors_climbed: typing.Optional[int] = dataclasses.field(default=None)
    swimming: SwimmingSummary = dataclasses.field(default_factory=SwimmingSummary)
    elevation: ElevationSummary = dataclasses.field(default_factory=ElevationSummary)


@dataclasses.dataclass
class DistanceDataDetailed(base_model.TerraDataModel):
    step_samples: typing.List[samples_.StepSample] = dataclasses.field(default_factory=list)
    distance_samples: typing.List[samples_.DistanceSample] = dataclasses.field(default_factory=list)
    elevation_samples: typing.List[samples_.ElevationSample] = dataclasses.field(default_factory=list)
    floors_climbed_samples: typing.List[samples_.FloorsClimbedSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class DistanceData(base_model.TerraDataModel):
    summary: DistanceDataSummary = dataclasses.field(default_factory=DistanceDataSummary)
    detailed: DistanceDataDetailed = dataclasses.field(default_factory=DistanceDataDetailed)


@dataclasses.dataclass
class PositionData(base_model.TerraDataModel):
    start_pos_lat_lng_deg: typing.List[float] = dataclasses.field(default_factory=list)
    center_pos_lat_lng_deg: typing.List[float] = dataclasses.field(default_factory=list)
    end_pos_lat_lng_deg: typing.List[float] = dataclasses.field(default_factory=list)
    position_samples: typing.List[samples_.PositionSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class ActiveDurationsData(base_model.TerraDataModel):
    activity_seconds: typing.Optional[float] = dataclasses.field(default=None)
    inactivity_seconds: typing.Optional[float] = dataclasses.field(default=None)
    rest_seconds: typing.Optional[float] = dataclasses.field(default=None)
    low_intensity_seconds: typing.Optional[float] = dataclasses.field(default=None)
    moderate_intensity_seconds: typing.Optional[float] = dataclasses.field(default=None)
    vigorous_intensity_seconds: typing.Optional[float] = dataclasses.field(default=None)
    num_continuous_inactive_periods: typing.Optional[int] = dataclasses.field(default=None)
    activity_levels_samples: typing.List[samples_.ActivityLevelSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class METData(base_model.TerraDataModel):
    avg_level: typing.Optional[float] = dataclasses.field(default=None)
    num_inactive_minutes: typing.Optional[float] = dataclasses.field(default=None)
    num_low_intensity_minutes: typing.Optional[float] = dataclasses.field(default=None)
    num_moderate_intensity_minutes: typing.Optional[float] = dataclasses.field(default=None)
    num_high_intensity_minutes: typing.Optional[float] = dataclasses.field(default=None)
    MET_samples: typing.List[samples_.METSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class MovementData(base_model.TerraDataModel):
    avg_speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    max_speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    adjusted_max_speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    normalized_speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    avg_pace_minutes_per_kilometer: typing.Optional[float] = dataclasses.field(default=None)
    max_pace_minutes_per_kilometer: typing.Optional[float] = dataclasses.field(default=None)
    avg_velocity_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    max_velocity_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    avg_cadence_rpm: typing.Optional[float] = dataclasses.field(default=None)
    max_cadence_rpm: typing.Optional[float] = dataclasses.field(default=None)
    avg_torque_newton_meters: typing.Optional[float] = dataclasses.field(default=None)
    max_torque_newton_meters: typing.Optional[float] = dataclasses.field(default=None)
    cadence_samples: typing.List[samples_.CadenceSample] = dataclasses.field(default_factory=list)
    speed_samples: typing.List[samples_.SpeedSample] = dataclasses.field(default_factory=list)
    torque_samples: typing.List[samples_.TorqueSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class CaloriesData(base_model.TerraDataModel):
    net_activity_calories: typing.Optional[float] = dataclasses.field(default=None)
    BMR_calories: typing.Optional[float] = dataclasses.field(default=None)
    total_burned_calories: typing.Optional[float] = dataclasses.field(default=None)
    net_intake_calories: typing.Optional[float] = dataclasses.field(default=None)
    calorie_samples: typing.List[samples_.CalorieSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class WorkData(base_model.TerraDataModel):
    work_kilojoules: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class PowerData(base_model.TerraDataModel):
    avg_watts: typing.Optional[float] = dataclasses.field(default=None)
    max_watts: typing.Optional[float] = dataclasses.field(default=None)
    power_samples: typing.List[samples_.PowerSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class EnergyData(base_model.TerraDataModel):
    energy_kilojoules: typing.Optional[float] = dataclasses.field(default=None)
    energy_planned_kilojoules: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class TSSData(base_model.TerraDataModel):
    TSS_samples: typing.List[samples_.TSSSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class HeartRateZone(base_model.TerraDataModel):
    zone: typing.Optional[int] = dataclasses.field(default=None)
    start_percentage: typing.Optional[float] = dataclasses.field(default=None)
    end_percentage: typing.Optional[float] = dataclasses.field(default=None)
    name: typing.Optional[str] = dataclasses.field(default=None)
    duration_seconds: typing.Optional[float] = dataclasses.field(default=None)

    def __post_init__(self) -> None:
        if not any((self.start_percentage, self.end_percentage, self.name)):
            self.start_percentage, self.end_percentage, self.name = (None, None, None)


@dataclasses.dataclass
class HeartRateDataSummary(base_model.TerraDataModel):
    avg_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    max_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    min_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    avg_hrv_rmssd: typing.Optional[float] = dataclasses.field(default=None)
    avg_hrv_sdnn: typing.Optional[float] = dataclasses.field(default=None)
    user_max_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    resting_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    hr_zone_data: typing.List[HeartRateZone] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class HeartRateDataDetailed(base_model.TerraDataModel):
    hr_samples: typing.List[samples_.HeartRateDataSample] = dataclasses.field(default_factory=list)
    hrv_samples_rmssd: typing.List[samples_.HeartRateVariabilityDataSampleRMSSD] = dataclasses.field(
        default_factory=list
    )
    hrv_samples_sdnn: typing.List[samples_.HeartRateVariabilityDataSampleSDNN] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class HeartRateData(base_model.TerraDataModel):
    summary: HeartRateDataSummary = dataclasses.field(default_factory=HeartRateDataSummary)
    detailed: HeartRateDataDetailed = dataclasses.field(default_factory=HeartRateDataDetailed)


@dataclasses.dataclass
class OxygenData(base_model.TerraDataModel):
    vo2max_ml_per_min_per_kg: typing.Optional[float] = dataclasses.field(default=None)
    avg_saturation_percentage: typing.Optional[float] = dataclasses.field(default=None)
    saturation_samples: typing.List[samples_.OxygenSaturationSample] = dataclasses.field(default_factory=list)
    vo2_samples: typing.List[samples_.Vo2MaxSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class PolylineMapData(base_model.TerraDataModel):
    summary_polyline: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class DeviceData(base_model.TerraDataModel):
    name: typing.Optional[str] = dataclasses.field(default=None)
    activation_timestamp: typing.Optional[str] = dataclasses.field(default=None)
    manufacturer: typing.Optional[str] = dataclasses.field(default=None)
    serial_number: typing.Optional[str] = dataclasses.field(default=None)
    software_version: typing.Optional[str] = dataclasses.field(default=None)
    hardware_version: typing.Optional[str] = dataclasses.field(default=None)
    other_devices: typing.List[samples_.OtherDeviceData] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Activity(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    lap_data: LapData = dataclasses.field(default_factory=LapData)
    distance_data: DistanceData = dataclasses.field(default_factory=DistanceData)
    position_data: PositionData = dataclasses.field(default_factory=PositionData)
    active_durations_data: ActiveDurationsData = dataclasses.field(default_factory=ActiveDurationsData)
    MET_data: METData = dataclasses.field(default_factory=METData)
    movement_data: MovementData = dataclasses.field(default_factory=MovementData)
    calories_data: CaloriesData = dataclasses.field(default_factory=CaloriesData)
    work_data: WorkData = dataclasses.field(default_factory=WorkData)
    power_data: PowerData = dataclasses.field(default_factory=PowerData)
    energy_data: EnergyData = dataclasses.field(default_factory=EnergyData)
    TSS_data: TSSData = dataclasses.field(default_factory=TSSData)
    heart_rate_data: HeartRateData = dataclasses.field(default_factory=HeartRateData)
    strain_data: StrainData = dataclasses.field(default_factory=StrainData)
    oxygen_data: OxygenData = dataclasses.field(default_factory=OxygenData)
    polyline_map_data: PolylineMapData = dataclasses.field(default_factory=PolylineMapData)
    device_data: DeviceData = dataclasses.field(default_factory=DeviceData)
    cheat_detection: typing.Optional[float] = dataclasses.field(default=None)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2 import activity
from terra.models.v2 import samples as samples_
from terra.models.v2.activity import OxygenData

__all__ = ["Body"]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class ECGReading(base_model.TerraDataModel):
    start_timestamp: typing.Optional[str] = dataclasses.field(default=None)
    avg_hr_bpm: typing.Optional[float] = dataclasses.field(default=None)
    afib_classfication: typing.Optional[int] = dataclasses.field(default=None)
    raw_signal: typing.List[samples_.RawECGSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class DeviceData(base_model.TerraDataModel):
    name: typing.Optional[str] = dataclasses.field(default=None)
    activation_timestamp: typing.Optional[str] = dataclasses.field(default=None)
    manufacturer: typing.Optional[str] = dataclasses.field(default=None)
    serial_number: typing.Optional[str] = dataclasses.field(default=None)
    software_version: typing.Optional[str] = dataclasses.field(default=None)
    hardware_version: typing.Optional[str] = dataclasses.field(default=None)
    other_devices: typing.List[samples_.OtherDeviceData] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class MeasurementsData(base_model.TerraDataModel):
    measurements: typing.List[samples_.MeasurementDataSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class TemperatureData(base_model.TerraDataModel):
    ambient_temperature_samples: typing.List[samples_.TemperatureSample] = dataclasses.field(default_factory=list)
    body_temperature_samples: typing.List[samples_.TemperatureSample] = dataclasses.field(default_factory=list)
    skin_temperature_samples: typing.List[samples_.TemperatureSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class HydrationData(base_model.TerraDataModel):
    hydration_amount_samples: typing.List[samples_.HydrationMeasurementSample] = dataclasses.field(default_factory=list)
    day_total_water_consumption_ml: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartData(base_model.TerraDataModel):
    heart_rate_data: activity.HeartRateData = dataclasses.field(default_factory=activity.HeartRateData)
    pulse_wave_velocity_samples: typing.List[samples_.PulseVelocitySample] = dataclasses.field(default_factory=list)
    afib_classification_samples: typing.List[samples_.AFibClassificationSample] = dataclasses.field(
        default_factory=list
    )
    ecg_signal: typing.List[ECGReading] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class BloodPressureData(base_model.TerraDataModel):
    blood_pressure_samples: typing.List[samples_.BloodPressureSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class GlucoseData(base_model.TerraDataModel):
    day_avg_blood_glucose_mg_per_dL: typing.Optional[float] = dataclasses.field(default=None)
    blood_glucose_samples: typing.List[samples_.GlucoseDataSample] = dataclasses.field(default_factory=list)
    detailed_blood_glucose_samples: typing.List[samples_.GlucoseDataSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class KetoneData(base_model.TerraDataModel):
    ketone_samples: typing.List[samples_.KetoneDataSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Body(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    measurements_data: MeasurementsData = dataclasses.field(default_factory=MeasurementsData)
    temperature_data: TemperatureData = dataclasses.field(default_factory=TemperatureData)
    hydration_data: HydrationData = dataclasses.field(default_factory=HydrationData)
    oxygen_data: OxygenData = dataclasses.field(default_factory=OxygenData)
    heart_data: HeartData = dataclasses.field(default_factory=HeartData)
    blood_pressure_data: BloodPressureData = dataclasses.field(default_factory=BloodPressureData)
    glucose_data: GlucoseData = dataclasses.field(default_factory=GlucoseData)
    device_data: DeviceData = dataclasses.field(default_factory=DeviceData)
    ketone_data: KetoneData = dataclasses.field(default_factory=KetoneData)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
__all__ = ["TerraException", "NoDtypeException", "NoBodyException", "NoUserInfoException", "NoClientAvailable"]


class TerraException(Exception):
    """Base class for all exceptions raised by this library"""


class NoClientAvailable(TerraException):
    """
    Exception raised when a TerraUser object has no Terra client set,
    but a method requiring it is called.
    """


class NoBodyException(TerraException):
    """
    Exception raised when a TerraApi object has no body
    """


class NoUserInfoException(TerraException):
    """
    Exception raised when a UserInfoDict is None
    """


class NoDtypeException(TerraException):
    """
    Exception raised when a TerraPi object has no dtype
    """



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2 import samples as samples_

__all__ = ["Menstruation"]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class MenstruationData(base_model.TerraDataModel):
    period_start_date: typing.Optional[str] = dataclasses.field(default=None)
    day_in_cycle: typing.Optional[int] = dataclasses.field(default=None)
    period_length_days: typing.Optional[int] = dataclasses.field(default=None)
    current_phase: typing.Optional[int] = dataclasses.field(default=None)
    length_of_current_phase_days: typing.Optional[int] = dataclasses.field(default=None)
    days_until_next_phase: typing.Optional[int] = dataclasses.field(default=None)
    predicted_cycle_length_days: typing.Optional[int] = dataclasses.field(default=None)
    is_predicted_cycle: typing.Optional[str] = dataclasses.field(default=None)
    cycle_length_days: typing.Optional[str] = dataclasses.field(default=None)
    last_updated_time: typing.Optional[str] = dataclasses.field(default=None)
    menstruation_flow: typing.List[samples_.MenstruationFlowSample] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class Menstruation(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    menstruation_data: MenstruationData = dataclasses.field(default_factory=MenstruationData)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model

__all__ = [
    "GlucoseDataSample",
    "HeartRateDataSample",
    "LapSample",
    "HeartRateVariabilityDataSampleRMSSD",
    "HeartRateVariabilityDataSampleSDNN",
    "DistanceSample",
    "StepSample",
    "ElevationSample",
    "PositionSample",
    "PowerSample",
    "SpeedSample",
    "CadenceSample",
    "ActivityLevelSample",
    "METSample",
    "TSSSample",
    "SleepHypnogramSample",
    "OxygenSaturationSample",
    "BreathSample",
    "SnoringSample",
    "StressSample",
    "BloodPressureSample",
    "MeasurementDataSample",
    "BodyTemperatureSample",
    "SkinTemperatureSample",
    "AmbientTemperatureSample",
    "TemperatureSample",
    "OtherDeviceData",
    "HydrationMeasurementSample",
    "Vo2MaxSample",
    "PulseVelocitySample",
    "AFibClassificationSample",
    "DrinkSample",
]


@dataclasses.dataclass
class GlucoseDataSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    blood_glucose_mg_per_dL: typing.Optional[int] = dataclasses.field(default=None)
    glucose_level_flag: typing.Optional[int] = dataclasses.field(default=None)
    trend_arrow: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class KetoneDataSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    ketone_mg_per_dL: typing.Optional[float] = dataclasses.field(default=None)
    sample_type: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class RawECGSample(base_model.TerraDataModel):
    potential_uV: typing.Optional[float] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartRateDataSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    bpm: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class LapSample(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)
    distance_meters: typing.Optional[float] = dataclasses.field(default=None)
    calories: typing.Optional[float] = dataclasses.field(default=None)
    total_strokes: typing.Optional[float] = dataclasses.field(default=None)
    stroke_type: typing.Optional[str] = dataclasses.field(default=None)
    avg_hr_bpm: typing.Optional[str] = dataclasses.field(default=None)
    avg_speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartRateVariabilityDataSampleRMSSD(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    hrv_rmssd: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class HeartRateVariabilityDataSampleSDNN(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    hrv_sdnn: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class DistanceSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    distance_meters: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class StepSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    steps: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class CalorieSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    calories: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class FloorsClimbedSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    floors_climbed: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class ElevationSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    elev_meters: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class PositionSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    coords_lat_lng_deg: typing.List[float] = dataclasses.field(default_factory=list)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class PowerSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    watts: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class TorqueSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    torque_newton_meters: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SpeedSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    speed_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class CadenceSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    cadence_rpm: typing.Optional[float] = dataclasses.field(default=None)
    timer_duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class ActivityLevelSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    activity_level: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class METSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    level: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class TSSSample(base_model.TerraDataModel):
    planned: typing.Optional[float] = dataclasses.field(default=None)
    actual: typing.Optional[float] = dataclasses.field(default=None)
    method: typing.Optional[str] = dataclasses.field(default=None)
    intensity_factor_planned: typing.Optional[float] = dataclasses.field(default=None)
    intensity_factor_actual: typing.Optional[float] = dataclasses.field(default=None)
    normalized_power_watts: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SleepHypnogramSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    level: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class OxygenSaturationSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    percentage: typing.Optional[float] = dataclasses.field(default=None)
    type: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class BreathSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    breaths_per_min: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SnoringSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    duration_seconds: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class StressSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    level: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class BloodPressureSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    diastolic_bp: typing.Optional[float] = dataclasses.field(default=None)
    systolic_bp: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class MeasurementDataSample(base_model.TerraDataModel):
    measurement_time: typing.Optional[str] = dataclasses.field(default=None)
    BMI: typing.Optional[float] = dataclasses.field(default=None)
    BMR: typing.Optional[float] = dataclasses.field(default=None)
    RMR: typing.Optional[float] = dataclasses.field(default=None)
    estimated_fitness_age: typing.Optional[int] = dataclasses.field(default=None)
    skin_fold_mm: typing.Optional[float] = dataclasses.field(default=None)
    bodyfat_percentage: typing.Optional[float] = dataclasses.field(default=None)
    weight_kg: typing.Optional[float] = dataclasses.field(default=None)
    height_cm: typing.Optional[float] = dataclasses.field(default=None)
    bone_mass_g: typing.Optional[float] = dataclasses.field(default=None)
    muscle_mass_g: typing.Optional[float] = dataclasses.field(default=None)
    lean_mass_g: typing.Optional[float] = dataclasses.field(default=None)
    water_percentage: typing.Optional[float] = dataclasses.field(default=None)
    insulin_units: typing.Optional[float] = dataclasses.field(default=None)
    insulin_type: typing.Optional[str] = dataclasses.field(default=None)
    urine_color: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class BodyTemperatureSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    body_temperature_celsius: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class SkinTemperatureSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    skin_temperature_celsius: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class AmbientTemperatureSample(base_model.TerraDataModel):
    temperature_celsius: typing.Optional[float] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class TemperatureSample(base_model.TerraDataModel):
    temperature_celsius: typing.Optional[float] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class OtherDeviceData(base_model.TerraDataModel):
    name: typing.Optional[str] = dataclasses.field(default=None)
    manufacturer: typing.Optional[str] = dataclasses.field(default=None)
    serial_number: typing.Optional[str] = dataclasses.field(default=None)
    software_version: typing.Optional[str] = dataclasses.field(default=None)
    hardware_version: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class HydrationMeasurementSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    hydration_kg: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class Vo2MaxSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    vo2max_ml_per_min_per_kg: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class PulseVelocitySample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    pulse_wave_velocity_meters_per_second: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class AFibClassificationSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    afib_classification: typing.Optional[int] = dataclasses.field(default=None)


@dataclasses.dataclass
class DrinkSample(base_model.TerraDataModel):
    drink_unit: typing.Optional[str] = dataclasses.field(default=None)
    drink_volume: typing.Optional[str] = dataclasses.field(default=None)
    drink_name: typing.Optional[str] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class MenstruationFlowSample(base_model.TerraDataModel):
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    flow: typing.Optional[int] = dataclasses.field(default=None)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models.v2.samples import DrinkSample

__all__ = ["Nutrition"]


@dataclasses.dataclass
class Metadata(base_model.TerraDataModel):
    start_time: typing.Optional[str] = dataclasses.field(default=None)
    end_time: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class Macros(base_model.TerraDataModel):
    calories: typing.Optional[float] = dataclasses.field(default=None)
    protein_g: typing.Optional[float] = dataclasses.field(default=None)
    carbohydrates_g: typing.Optional[float] = dataclasses.field(default=None)
    fat_g: typing.Optional[float] = dataclasses.field(default=None)
    trans_fat_g: typing.Optional[float] = dataclasses.field(default=None)
    saturated_fat_g: typing.Optional[float] = dataclasses.field(default=None)
    sugar_g: typing.Optional[float] = dataclasses.field(default=None)
    cholesterol_mg: typing.Optional[float] = dataclasses.field(default=None)
    fiber_g: typing.Optional[float] = dataclasses.field(default=None)
    sodium_mg: typing.Optional[float] = dataclasses.field(default=None)
    alcohol_g: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class Micros(base_model.TerraDataModel):
    biotin_mg: typing.Optional[float] = dataclasses.field(default=None)
    caffeine_mg: typing.Optional[float] = dataclasses.field(default=None)
    chloride_mg: typing.Optional[float] = dataclasses.field(default=None)
    chromium_mg: typing.Optional[float] = dataclasses.field(default=None)
    copper_mg: typing.Optional[float] = dataclasses.field(default=None)
    calcium_mg: typing.Optional[float] = dataclasses.field(default=None)
    folate_mg: typing.Optional[float] = dataclasses.field(default=None)
    folic_acid_mg: typing.Optional[float] = dataclasses.field(default=None)
    iodine_mg: typing.Optional[float] = dataclasses.field(default=None)
    iron_mg: typing.Optional[float] = dataclasses.field(default=None)
    magnesium_mg: typing.Optional[float] = dataclasses.field(default=None)
    manganese_mg: typing.Optional[float] = dataclasses.field(default=None)
    molybdenum_mg: typing.Optional[float] = dataclasses.field(default=None)
    monounsaturated_fat_g: typing.Optional[float] = dataclasses.field(default=None)
    niacin_mg: typing.Optional[float] = dataclasses.field(default=None)
    pantothenic_acid_mg: typing.Optional[float] = dataclasses.field(default=None)
    phosphorus_mg: typing.Optional[float] = dataclasses.field(default=None)
    polyunsaturated_fat_g: typing.Optional[float] = dataclasses.field(default=None)
    potassium_mg: typing.Optional[float] = dataclasses.field(default=None)
    riboflavin_mg: typing.Optional[float] = dataclasses.field(default=None)
    selenium_mg: typing.Optional[float] = dataclasses.field(default=None)
    thiamin_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_A_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_B12_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_B6_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_C_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_D_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_E_mg: typing.Optional[float] = dataclasses.field(default=None)
    vitamin_K_mg: typing.Optional[float] = dataclasses.field(default=None)
    zinc_mg: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class Quantity(base_model.TerraDataModel):
    unit: typing.Optional[int] = dataclasses.field(default=None)
    amount: typing.Optional[float] = dataclasses.field(default=None)


@dataclasses.dataclass
class Meal(base_model.TerraDataModel):
    name: typing.Optional[str] = dataclasses.field(default=None)
    id: typing.Optional[str] = dataclasses.field(default=None)
    timestamp: typing.Optional[str] = dataclasses.field(default=None)
    type: typing.Optional[int] = dataclasses.field(default=0)
    quantity: Quantity = dataclasses.field(default_factory=Quantity)
    macros: Macros = dataclasses.field(default_factory=Macros)
    micros: Micros = dataclasses.field(default_factory=Micros)


@dataclasses.dataclass
class NutritionSummary(base_model.TerraDataModel):
    water_ml: typing.Optional[float] = dataclasses.field(default=None)
    macros: Macros = dataclasses.field(default_factory=Macros)
    drink_ml: typing.Optional[float] = dataclasses.field(default=None)
    micros: Micros = dataclasses.field(default_factory=Micros)


@dataclasses.dataclass
class Nutrition(base_model.TerraDataModel):
    metadata: Metadata = dataclasses.field(default_factory=Metadata)
    summary: NutritionSummary = dataclasses.field(default_factory=NutritionSummary)
    meals: typing.List[Meal] = dataclasses.field(default_factory=list)
    drink_samples: typing.List[DrinkSample] = dataclasses.field(default_factory=list)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import re
import types

from setuptools import find_namespace_packages
from setuptools import setup

name = "terra"


def parse_meta():
    with open(os.path.join(name, "__init__.py")) as fp:
        code = fp.read()

    token_pattern = re.compile(r"^__(?P<key>\w+)?__\s*=\s*(?P<quote>(?:'{3}|\"{3}|'|\"))(?P<value>.*?)(?P=quote)", re.M)

    groups = {}

    for match in token_pattern.finditer(code):
        group = match.groupdict()
        groups[group["key"]] = group["value"]

    return types.SimpleNamespace(**groups)


def long_description():
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path):
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]


meta = parse_meta()

setup(
    name="terra-python",
    version=meta.version,
    description="A python wrapper for the Terra API",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author="Terra Enabling Developers",
    author_email="dev@tryterra.co",
    url="https://github.com/tryterra/terra-client-python",
    packages=find_namespace_packages(include=[name + "*"]),
    license="APACHE-2.0-ONLY",
    include_package_data=True,
    zip_safe=False,
    install_requires=parse_requirements_file("requirements.txt"),
    python_requires=">=3.8.0,<3.11",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: Apache Software License"
    ],
)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import dataclasses
import typing

from terra.models import base_model
from terra.models import user as user_

__all__ = ["Athlete", "AthleteCollection"]


@dataclasses.dataclass
class Athlete(base_model.TerraDataModel):
    first_name: typing.Optional[str] = dataclasses.field(default=None)
    last_name: typing.Optional[str] = dataclasses.field(default=None)
    gender: typing.Optional[str] = dataclasses.field(default=None)
    sex: typing.Optional[str] = dataclasses.field(default=None)
    age: typing.Optional[int] = dataclasses.field(default=None)
    date_of_birth: typing.Optional[str] = dataclasses.field(default=None)
    bio: typing.Optional[str] = dataclasses.field(default=None)
    email: typing.Optional[str] = dataclasses.field(default=None)
    city: typing.Optional[str] = dataclasses.field(default=None)
    state: typing.Optional[str] = dataclasses.field(default=None)
    country: typing.Optional[str] = dataclasses.field(default=None)


@dataclasses.dataclass
class AthleteCollection(base_model.TerraDataModel):
    user: user_.User = dataclasses.field()
    athlete: Athlete = dataclasses.field()
    type: typing.Optional[str] = dataclasses.field(default=None)



#  Copyright 2022 Terra Enabling Developers Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

__all__ = [
    "utils",
    "exceptions",
    "constants",
    "base_client",
    "models",
    "api",
    "Terra",
    "TerraException",
    "NoBodyException",
    "NoClientAvailable",
    "NoDtypeException",
    "NoUserInfoException",
]

from terra import api
from terra import base_client
from terra import constants
from terra import exceptions
from terra import models
from terra import utils
from terra.base_client import *
from terra.exceptions import *

__version__ = "0.0.8"
